/************************************************************************
 *                                                                      *
 *  AllocP - supply a better AllocMem/AllocVec                          *
 *                                                                      *
 *  written by Andreas R. Kleinert                                      *
 *  Andreas_Kleinert@t-online.de                                        *
 *                                                                      *
 *  No Assembler - 100 percent C                                        *
 *                                                                      *
 * Hacked (unofficial!) :                                               *
 *        - 4k page alignment                                           *
 *
 * V2.0 :                                                               *
 *        - 32 Byte alignment                                           *
 *                                                                      *
 * V1.5 :                                                               *
 *        - oops, Exec's alignment already is 8                         *
 *          (stay tuned, I'd like to write a 32 alignment patch,        *
 *           having PPC in mind)                                        *
 *                                                                      *
 * V1.4 :                                                               *
 *        - fixed $version                                              *
 *        - fixed self-detaching (no more CON)                          *
 *                                                                      *
 * V1.3 :                                                               *
 *        - when the real "avail flush" process took place,             *
 *          a memory loss could have occured                            *
 *        - added Forbid/Permit                                         *
 *        - fixed/changed some other stuff                              *
 *                                                                      *
 * V1.2 :                                                               *
 *        - redone with SAS/C 6.58                                      *
 *        - now checks for MEMF_NO_EXPUNGE                              *
 *        - utility.library was not needed                              *
 *        - fixed requester layout                                      *
 *        - fixed #include list                                         *
 *        - now 68020 compiled (040 before, thus 020 ;-)                *
 *        - fixed $VERsion string                                       *
 *        - etc.                                                        *
 *        - saved > 400 Bytes                                           *
 *                                                                      *
 * V1.1 :                                                               *
 *        - V37 works, too                                              *
 *        - alignment feature was broken                                *
 *        - reduced stack size                                          *
 *                                                                      *
 * V1.0 :                                                               *
 *        - first release                                               *
 *                                                                      *
 ************************************************************************/

#define __USE_SYSBASE

#define SAFE
/* #define SAFEST */

#include <exec/execbase.h>
#include <exec/memory.h>
#include <intuition/intuition.h>
#include <libraries/iffparse.h>

#include <string.h>
#include <stdio.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/intuition.h>

#define N (NULL)


extern struct ExecBase *SysBase;

struct IntuitionBase *IntuitionBase = N;


#define NEWMEMID MAKE_ID('B','E','T','A')

struct NewMemStr
{
 ULONG  code;
 UBYTE *fullptr;
 UBYTE *useptr;
 ULONG  size;
 ULONG  origsize;
 ULONG  origsize4;
}; /* 24 Bytes + 8-alignment makes 32 */

APTR __far NewAllocMemVec;
ULONG __far __asm (*OrigAllocMemVec)(register __d0 ULONG size,
                                     register __d1 ULONG memf,
                                     register __a6 struct ExecBase *ExecBase_a6);

APTR __far NewAllocVecVec;
ULONG __far __asm (*OrigAllocVecVec)(register __d0 ULONG size,
                                     register __d1 ULONG memf,
                                     register __a6 struct ExecBase *ExecBase_a6);

APTR __far NewFreeMemVec;
ULONG __far __asm (*OrigFreeMemVec)(register __a1 UBYTE *mptr,
                                    register __d0 ULONG size,
                                    register __a6 struct ExecBase *ExecBase_a6);

APTR __far NewFreeVecVec;
ULONG __far __asm (*OrigFreeVecVec)(register __a1 UBYTE *mptr,
                                    register __a6 struct ExecBase *ExecBase_a6);


ULONG __asm NewAllocMem(register __d0 ULONG size,
                        register __d1 ULONG memf,
                        register __a6 struct ExecBase *ExecBase_a6)
{
 ULONG newsize;
 struct NewMemStr *ptr;
 UBYTE * fullptr;
 UBYTE * useptr;


 if(!size) return(N);

 newsize = size + sizeof(struct NewMemStr) + 4096;

 if(newsize < size) newsize = size; /* flush call */

          ptr = (APTR) OrigAllocMemVec(newsize, memf, ExecBase_a6);
 if(!ptr) ptr = (APTR) OrigAllocMemVec(newsize, memf, ExecBase_a6);

 if(ptr)
  {
   fullptr = (UBYTE *) ptr;	  
   useptr     = (UBYTE *)((ULONG)(fullptr + sizeof(struct NewMemStr) + 4096)&0xfffff000ul);

   ptr = (APTR) (useptr - sizeof(struct NewMemStr));
   ptr->code       = NEWMEMID;
   ptr->fullptr    = fullptr;
   ptr->useptr     = useptr;
   ptr->size       = newsize;
   ptr->origsize   = size;
   ptr->origsize4  = ptr->origsize + 4; /* imitate AllocVec */

   return( (ULONG) ptr->useptr );
  }

 return( (ULONG) N);
}

ULONG __asm NewAllocVec(register __d0 ULONG size,
                        register __d1 ULONG memf,
                        register __a6 struct ExecBase *ExecBase_a6)
{
 ULONG newsize;
 struct NewMemStr *ptr;
 UBYTE * fullptr;
 UBYTE * useptr;


 if(!size) return(N);

 newsize = size + sizeof(struct NewMemStr)+4096;

 if(newsize < size) newsize = size; /* flush call */

          ptr = (APTR) OrigAllocMemVec(newsize, memf, ExecBase_a6);
 if(!ptr) ptr = (APTR) OrigAllocMemVec(newsize, memf, ExecBase_a6);

 if(ptr)
  {
   fullptr = (UBYTE *) ptr;	  
   useptr     = (UBYTE *)((ULONG)(fullptr + sizeof(struct NewMemStr) + 4096)&0xfffff000ul);

   ptr = (APTR) (useptr - sizeof(struct NewMemStr));
   ptr->code       = NEWMEMID;
   ptr->fullptr    = fullptr;
   ptr->useptr     = useptr;
   ptr->size       = newsize;
   ptr->origsize   = size;
   ptr->origsize4  = ptr->origsize + 4; /* imitate AllocVec */

   return( (ULONG) ptr->useptr );
  }

 return( (ULONG) N);
}

void __asm NewFreeMem(register __a1 UBYTE *mptr,
                      register __d0 ULONG size,
                      register __a6 struct ExecBase *ExecBase_a6)
{
 struct NewMemStr *ptr = (struct NewMemStr *) (mptr - sizeof(struct NewMemStr));

 if(ptr->code == NEWMEMID)                    /* first check  */
  {
#ifdef SAFE
   if(ptr->fullptr < ptr->useptr)             /* second check */
    {
#ifdef SAFEST
     if(ptr->size == ptr->origsize)           /* third check  */
#endif /* SAFEST */
#endif /* SAFE   */
      {
       OrigFreeMemVec(ptr->fullptr, ptr->size, ExecBase_a6);
       return;
      }
    }
  }

 OrigFreeMemVec(mptr, size, ExecBase_a6);
}

void __asm NewFreeVec(register __a1 UBYTE *mptr,
                      register __a6 struct ExecBase *ExecBase_a6)
{
 struct NewMemStr *ptr = (struct NewMemStr *) (mptr - sizeof(struct NewMemStr));

 if(ptr->code == NEWMEMID)                    /* first check  */
  {
#ifdef SAFE
   if(ptr->fullptr < ptr->useptr)             /* second check */
    {
#ifdef SAFEST
     if(ptr->origsize == (ptr->origsize + 4)) /* third check  */
#endif /* SAFEST */
#endif /* SAFE   */
      {
       OrigFreeMemVec(ptr->fullptr, ptr->size, ExecBase_a6);
       return;
      }
    }
  }

 OrigFreeVecVec(mptr, ExecBase_a6);
}

/* *************************************************** */
/* *                                                 * */
/* * Compiler Stuff for BackgroundIO                 * */
/* *                                                 * */
/* *************************************************** */

long  __stack        = 1024;
char *__procname     = "AllocP";
long  __priority     = 1;
long  __BackGroundIO = 0;

void __regargs __chkabort(void) { }
void __regargs _CXBRK(void)     { }

char vertext [] = "\0$VER: AllocP Hacked (4.1.22)";


long main(long argc, char **argv)
{
 long retval = 0;

 APTR task;

 Forbid();
 task = FindTask(N);
 Permit();

 if(task) SetTaskPri(task, __priority);


 IntuitionBase = (APTR) OpenLibrary("intuition.library", 37);
 if(IntuitionBase)
  {
   AllocMem(0xFFFFFFFF, 0L); /* flush first                         */
   AllocMem(0xFFFFFFFF, 0L); /* (less "old" allocations left, then) */

   Forbid();
   OrigAllocMemVec = (long (* __asm )(register __d0 ULONG size,
                                      register __d1 ULONG memf,
                                      register __a6 struct ExecBase *ExecBase_a6)) SetFunction((APTR)SysBase, -0xc6, (APTR)NewAllocMem);

   OrigAllocVecVec = (long (* __asm )(register __d0 ULONG size,
                                      register __d1 ULONG memf,
                                      register __a6 struct ExecBase *ExecBase_a6)) SetFunction((APTR)SysBase, -0x2ac, (APTR)NewAllocVec);

   OrigFreeMemVec = (long (* __asm )(register __a1 UBYTE *mptr,
                                     register __d0 ULONG size,
                                     register __a6 struct ExecBase *ExecBase_a6)) SetFunction((APTR)SysBase, -0xd2, (APTR)NewFreeMem);

   OrigFreeVecVec = (long (* __asm )(register __d1 UBYTE *mptr,
                                     register __a6 struct ExecBase *ExecBase_a6)) SetFunction((APTR)SysBase, -0x2b2, (APTR)NewFreeVec);
   Permit();

   if(OrigAllocMemVec || OrigAllocVecVec)
    {
     if(task) SetTaskPri(task, -10);
     Wait(SIGBREAKF_CTRL_C);
     if(task) SetTaskPri(task, 1);

     Forbid();
     NewAllocMemVec = (APTR) SetFunction((APTR)SysBase, -0xc6,  (APTR)OrigAllocMemVec);
     NewAllocVecVec = (APTR) SetFunction((APTR)SysBase, -0x2ac, (APTR)OrigAllocVecVec);
     NewFreeMemVec  = (APTR) SetFunction((APTR)SysBase, -0xd2,  (APTR)OrigFreeMemVec);
     NewFreeVecVec  = (APTR) SetFunction((APTR)SysBase, -0x2b2, (APTR)OrigFreeVecVec);
     Permit();

     if(    (NewAllocMemVec != NewAllocMem)
         || (NewAllocVecVec != NewAllocVec)
         || (NewFreeMemVec  != NewFreeMem)
         || (NewFreeVecVec  != NewFreeVec)  )
      {
       ULONG idcmp = N;
       struct EasyStruct estr;

       estr.es_StructSize   = sizeof(struct EasyStruct);
       estr.es_Flags        = N;
       estr.es_Title        = "AllocP Request";
       estr.es_TextFormat   = "Error while removing from system !\n"
                              "There were more patches !\n"
                              "Do a reset soon !";
       estr.es_GadgetFormat = "Ok";

       EasyRequestArgs(N, &estr, (ULONG *) &idcmp, N);
      }

    }else retval = 20;

   CloseLibrary((struct Library *) IntuitionBase);

  }else retval = 20;

 return(retval);
}
